/*
 *   FreeM - Database Server
 * 
 *   2001-OCT-30 Axel Trocha:  inital version
 *   2001-NOV-22 Axel Trocha:  mostly rewritten
 *
 */

#include <stdio.h>
#include <netinet/in.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>

#include "m_def0.h"

#define MAXCONNECTIONS 2


void refresh(int x)
{
   if ( fork() > 0 ) exit(0);
   signal (SIGXCPU, refresh);
}


struct user_s {
  int sock;
  char realident[64];
} user[MAXCONNECTIONS];


void userinit(void)
{
  int i;

  for ( i = 0; i < MAXCONNECTIONS; i++ )
    user[i].sock = 0;
}


int userfindfree(void)
{
  int i;

  for ( i = 0; i < MAXCONNECTIONS; i++ )
    if ( ! user[i].sock )
      return i;

  return -1;
}


int useradd(int sock)
{
  int new;

  if ((new = userfindfree()) == -1) return -1;
  user[new].sock = sock;
  *user[new].realident = '\0';

  return new;
}


void userdel(int sock)
{
  int i;

  for ( i = 0; i<MAXCONNECTIONS; i++ )
    if (user[i].sock == sock)
    {
      shutdown (sock, 2);
      close (sock);
      user[i].sock = 0;
    }
}
 

static int mess(int fd, const char *fmt, ...)
/*
 *  send a message to the chosen file-descriptor <fd>
 */
{
    char buffer[4000];
    va_list args;
    int l;

    va_start(args, fmt);
    l = vsnprintf(buffer, sizeof(buffer), fmt, args);
    va_end(args);

    if (l == -1) {
        /* for older libcs */
        l = sizeof(buffer);
    }
    return write(fd, buffer, l);
}


int m_parse (int fd, char *x,int len)
/*
 *   parse incoming data and serve it
 */
{
  switch (*x) {

    case DB_SERV_QUIT:
	userdel (fd);
	return -1;

    case DB_SERV_GET:
	{
	char node[256];
	char *codptr;
	char data[256];

	strncpy (node, x+1, len-3);
	node[len-3] = '\0';

/*
	mess (fd, "DB_GET %s.\n", node);

	codptr = (char *) node;
	expr (STRING);

	if (node[0] == '^') {
          mess (fd, "DB_QUERY %s.\n",codptr);
	  global (get_sym, node, data);
	}

*/
	mumps_get_var (node, data);
	
	//data[stlen(data)] = '\0';
	mess (fd, "-%s\n", data);

	}
	break;

    case DB_SERV_SET:
	mess (fd, "DB_WRITE\n");
	break;
  }
}


int m_server(int port)
{
  struct sockaddr_in net_name;
  int hangsock;
  fd_set descr1, descr2;
  struct hostent *hent;

  signal (SIGXCPU, refresh);

  printf ("mserver starting up...\n");

  if ((hangsock = socket (PF_INET, SOCK_STREAM, 0))<0)
  {
    fprintf (stderr, "Cannot open socket\n");
    return (-1);
  }

  net_name.sin_family = AF_INET;
  net_name.sin_addr.s_addr = INADDR_ANY;
  net_name.sin_port = htons (port);

  /*
   *   bind and listen to <port>
   */
  if (bind(hangsock, (struct sockaddr *)&net_name, sizeof(struct sockaddr_in)))
  {
    fprintf (stderr, "cannot bind port %d\n", port);
    fprintf (stderr, "exiting...\n\n");
    return (-1);
  }
  printf ("listening on port %d\n", port);
  printf ("awaiting connections...\n\n");
  listen(hangsock, MAXCONNECTIONS);

  switch (fork())
  {
      case 0:
           if (setsid() != -1) break;
      case -1:
           fprintf (stderr, "cannot set to background\n");
           return -1;
      default:
           return 0;
  }

  while (1)
  {
    int i;
    int n = 0;
    int max = 0;

    FD_ZERO(&descr1);
    FD_ZERO(&descr2);

    for (i=0; i<MAXCONNECTIONS; i++)
      if (user[i].sock > 0)
      {
	 n++;
         FD_SET(user[i].sock, &descr1);
         FD_SET(user[i].sock, &descr2);
         if (user[i].sock>max) max = user[i].sock;
      }

    FD_SET(hangsock, &descr1);
    if (hangsock>max) max = hangsock;

    n = select (max+1, &descr1, NULL, &descr2, NULL);
    {
      int x, saddrlen;

      saddrlen = sizeof(struct sockaddr_in);
      if ( FD_ISSET(hangsock, &descr1) )
      {
           struct sockaddr_in tmpaddr;
           int new;

           if ((new=accept(hangsock, (struct sockaddr *)&tmpaddr, &saddrlen))>0)
	     if (useradd (new) < 0)
	     {
               shutdown (new,2);
               close (new); }
	     else {
	     }
		
      }

      for (x=0; x < MAXCONNECTIONS; x++)
        if (user[x].sock > 0)
          if ( FD_ISSET(user[x].sock, &descr2) ) userdel(user[x].sock);
          else
            if ( FD_ISSET(user[x].sock, &descr1) )
            {
	      int rd;
	      char intbuff[4096];
	      int fd = user[x].sock;

              if ((rd = read (fd, intbuff, sizeof(intbuff)))>0)
              {
		m_parse(fd, intbuff, rd);
	      }
              else userdel (fd);
            }
    }
  }
}
